@@ -16,11 +16,30 @@ PageState.prototype = {
     init: function() {
         this.allowRemoteImages = false;
         this.isLoaded = false;
+        this.undoEnabled = false;
+        this.redoEnabled = false;
         this.hasSelection = false;
         this.lastPreferredHeight = 0;
 
         let state = this;
 
+        // Set up an observer to keep track of modifications made to
+        // the document when editing.
+        let modifiedId = null;
+        this.bodyObserver = new MutationObserver(function(records) {
+            if (modifiedId == null) {
+                modifiedId = window.setTimeout(function() {
+                    state.documentModified();
+                    state.checkCommandStack();
+                    modifiedId = null;
+                }, 1000);
+            }
+        });
+
+        document.addEventListener("DOMContentLoaded", function(e) {
+            state.loaded();
+        });
+
         // Coalesce multiple calls to updatePreferredHeight using a
         // timeout to avoid the overhead of multiple JS messages sent
         // to the app and hence view multiple resizes being queued.
@@ -30,17 +49,10 @@ PageState.prototype = {
                 clearTimeout(queueTimeout);
             }
             queueTimeout = setTimeout(
-                function() { state.updatePreferredHeight(); }, 10
+                function() { state.updatePreferredHeight(); }, 100
             );
         };
 
-        // Queues an update after the DOM has been initially loaded
-        // and had any changes made to it by derived classes.
-        document.addEventListener("DOMContentLoaded", function(e) {
-            state.loaded();
-            queuePreferredHeightUpdate();
-        });
-
         // Queues an update when the complete document is loaded.
         //
         // Note also that the delay introduced here by this last call
@@ -57,22 +69,31 @@ PageState.prototype = {
         document.addEventListener("load", function(e) {
             queuePreferredHeightUpdate();
         }, true); // load does not bubble
+
+        // Queues an update if the window changes size, e.g. if the
+        // user resized the window
+        window.addEventListener("resize", function(e) {
+            queuePreferredHeightUpdate();
+        }, false); // load does not bubble
+
+        // Queues an update when a transition has completed, e.g. if the
+        // user resized the window
+        window.addEventListener("transitionend", function(e) {
+            queuePreferredHeightUpdate();
+        }, false); // load does not bubble
     },
     getPreferredHeight: function() {
-        let html = window.document.documentElement;
-        let height = html.offsetHeight;
-        let computed = window.getComputedStyle(html);
-        let top = computed.getPropertyValue('margin-top');
-        let bot = computed.getPropertyValue('margin-bottom');
-
-        return (
-            height
-                + parseInt(top.substring(0, top.length - 2))
-                + parseInt(bot.substring(0, bot.length - 2))
-        );
+        return window.document.body.scrollHeight;
+    },
+    getHtml: function() {
+        return document.body.innerHTML;
     },
     loaded: function() {
         this.isLoaded = true;
+        // Always fire a prefered height update first so that it will
+        // be vaguegly correct when notifying of the HTML load
+        // completing.
+        this.updatePreferredHeight();
         window.webkit.messageHandlers.contentLoaded.postMessage(null);
     },
     loadRemoteImages: function() {
@@ -85,6 +106,29 @@ PageState.prototype = {
             img.src = src;
         }
     },
+    setEditable: function(enabled) {
+        if (!enabled) {
+            this.stopBodyObserver();
+        }
+        document.body.contentEditable = enabled;
+        if (enabled) {
+            // Enable modification observation only after the document
+            // has been set editable as WebKit will alter some attrs
+            this.startBodyObserver();
+        }
+    },
+    startBodyObserver: function() {
+        let config = {
+            attributes: true,
+            childList: true,
+            characterData: true,
+            subtree: true
+        };
+        this.bodyObserver.observe(document.body, config);
+    },
+    stopBodyObserver: function() {
+        this.bodyObserver.disconnect();
+    },
     remoteImageLoadBlocked: function() {
         window.webkit.messageHandlers.remoteImageLoadBlocked.postMessage(null);
     },
@@ -96,15 +140,34 @@ PageState.prototype = {
             height = this.getPreferredHeight();
         }
 
-        let updated = false;
-        if (height > 0 && height != this.lastPreferredHeight) {
+        // Don't send the message until after the DOM has been fully
+        // loaded and processed by any derived classes. Since
+        // ConversationPageState may collapse any quotes, sending the
+        // current preferred height before then may send a value that
+        // is too large, causing the message body view to grow then
+        // shrink again, leading to visual flicker.
+        if (this.isLoaded && height > 0 && height != this.lastPreferredHeight) {
             updated = true;
             this.lastPreferredHeight = height;
             window.webkit.messageHandlers.preferredHeightChanged.postMessage(
                 height
             );
         }
-        return updated;
+    },
+    checkCommandStack: function() {
+        let canUndo = document.queryCommandEnabled("undo");
+        let canRedo = document.queryCommandEnabled("redo");
+
+        if (canUndo != this.undoEnabled || canRedo != this.redoEnabled) {
+            this.undoEnabled = canUndo;
+            this.redoEnabled = canRedo;
+            window.webkit.messageHandlers.commandStackChanged.postMessage(
+                this.undoEnabled + "," + this.redoEnabled
+            );
+        }
+    },
+    documentModified: function(element) {
+        window.webkit.messageHandlers.documentModified.postMessage(null);
     },
     selectionChanged: function() {
         let hasSelection = !window.getSelection().isCollapsed;
@@ -114,3 +177,5 @@ PageState.prototype = {
         }
     }
 };
+
+var geary = new PageState();
